#!/usr/bin/env python3
"""
Advanced Port Scanner
Feature-rich port scanner with service detection and OS fingerprinting.
Parameters:
    - target: Target IP or hostname
    - ports: Port range to scan (e.g., '1-1024' or '80,443,8080')
    - mode: Scan mode (syn, connect, udp)
Requires: scapy package and admin/root privileges for SYN and UDP scans
"""
from scapy.all import *
import socket
import concurrent.futures
import argparse
import logging
from datetime import datetime
import re
import json

class AdvancedPortScanner:
    def __init__(self, target, ports='1-1024', mode='syn', threads=50):
        self.target = target
        self.ports = self.parse_ports(ports)
        self.mode = mode.lower()
        self.threads = threads
        self.results = []
        self.setup_logging()

    def setup_logging(self):
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s'
        )
        self.logger = logging.getLogger(__name__)

    def parse_ports(self, ports):
        port_list = []
        for part in ports.split(','):
            if '-' in part:
                start, end = map(int, part.split('-'))
                port_list.extend(range(start, end + 1))
            else:
                port_list.append(int(part))
        return sorted(list(set(port_list)))

    def syn_scan(self, port):
        try:
            syn_packet = IP(dst=self.target)/TCP(dport=port, flags="S")
            response = sr1(syn_packet, timeout=2, verbose=0)
            
            if response and response.haslayer(TCP):
                flags = response[TCP].flags
                if flags & 0x12:  # SYN-ACK
                    service = self.identify_service(port)
                    return {
                        'port': port,
                        'state': 'open',
                        'service': service,
                        'method': 'SYN'
                    }
            return None
        except Exception as e:
            self.logger.error(f"Error scanning port {port}: {str(e)}")
            return None

    def connect_scan(self, port):
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(2)
            result = sock.connect_ex((self.target, port))
            
            if result == 0:
                service = self.identify_service(port)
                return {
                    'port': port,
                    'state': 'open',
                    'service': service,
                    'method': 'Connect'
                }
            return None
        except Exception as e:
            self.logger.error(f"Error scanning port {port}: {str(e)}")
            return None
        finally:
            sock.close()

    def udp_scan(self, port):
        try:
            udp_packet = IP(dst=self.target)/UDP(dport=port)
            response = sr1(udp_packet, timeout=2, verbose=0)
            
            if response is None:
                return {
                    'port': port,
                    'state': 'open|filtered',
                    'service': self.identify_service(port),
                    'method': 'UDP'
                }
            elif response.haslayer(ICMP):
                if int(response[ICMP].type) == 3 and int(response[ICMP].code) == 3:
                    return None  # Port closed
            return None
        except Exception as e:
            self.logger.error(f"Error scanning port {port}: {str(e)}")
            return None

    def identify_service(self, port):
        try:
            service = socket.getservbyport(port)
            return service
        except:
            return "unknown"

    def scan_port(self, port):
        if self.mode == 'syn':
            return self.syn_scan(port)
        elif self.mode == 'connect':
            return self.connect_scan(port)
        elif self.mode == 'udp':
            return self.udp_scan(port)

    def os_fingerprint(self):
        try:
            ttl_packet = IP(dst=self.target)/ICMP()
            response = sr1(ttl_packet, timeout=2, verbose=0)
            
            if response:
                ttl = response.ttl
                if ttl <= 64:
                    return 'Linux/Unix'
                elif ttl <= 128:
                    return 'Windows'
                else:
                    return 'Unknown'
        except:
            return 'Unknown'

    def scan(self):
        self.logger.info(f"Starting {self.mode.upper()} scan of {self.target}")
        self.logger.info(f"Scanning {len(self.ports)} ports...")

        with concurrent.futures.ThreadPoolExecutor(max_workers=self.threads) as executor:
            future_to_port = {executor.submit(self.scan_port, port): port 
                            for port in self.ports}
            
            for future in concurrent.futures.as_completed(future_to_port):
                result = future.result()
                if result:
                    self.results.append(result)

        os_guess = self.os_fingerprint()
        
        scan_results = {
            'target': self.target,
            'timestamp': datetime.now().isoformat(),
            'scan_type': self.mode,
            'os_guess': os_guess,
            'open_ports': sorted(self.results, key=lambda x: x['port'])
        }

        self.logger.info(f"\nScan Results for {self.target}")
        self.logger.info(f"OS Detection: {os_guess}")
        self.logger.info("\nOpen Ports:")
        
        for result in sorted(self.results, key=lambda x: x['port']):
            self.logger.info(
                f"Port {result['port']}/{self.mode}: {result['state']} "
                f"({result['service']})"
            )

        return scan_results

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Advanced Port Scanner")
    parser.add_argument("target", help="Target IP or hostname")
    parser.add_argument("--ports", default="1-1024",
                      help="Port range (e.g., '1-1024' or '80,443,8080')")
    parser.add_argument("--mode", choices=['syn', 'connect', 'udp'],
                      default='syn', help="Scan mode")
    parser.add_argument("--threads", type=int, default=50,
                      help="Number of threads")
    args = parser.parse_args()

    scanner = AdvancedPortScanner(args.target, args.ports, args.mode, args.threads)
    scanner.scan()
